//***********************************************************************//
//                                                                       //
//                               Direct Play                             //
//                                                                       //
//***********************************************************************//

#define WIN32_LEAN_AND_MEAN
 
#include <stdio.h>
#include <dplay.h>

//Smalltalk protocol:
//	open (guidString)
//	Use providerList () to prompt user for a choice.
//	Use chooseProvider (indexOfChoice) to pick the active provider.
//	Decide on one of the following:
//		a) To be a server
//			server (fullName, friendlyName)
//		b) To be a client
//			Use sessionList (timeout) to prompt user for a choice of sessions
//			Use chooseSession (indexOfChoice) to pick one.
//			client (fullName, friendlyName)
//	Use playerList () to get players in selected session.
//	Use choosePlayer (indexOfChoice) to pick the receiver for all subsequent sends.
//	Use receive/send to communicate
//	close
//
//	Obtaining an empty list suggests an error which can be queried via
//	lastError.

//***********************************************************************//
//                Public Prototypes and Forward References               //
//***********************************************************************//

typedef BYTE Boolean;
char *lastError ();

void openDirectPlay (char *guidString);
void closeDirectPlay ();

char *providerList ();
void chooseProvider (long choice);

char *sessionList (long timeout);
void chooseSession (long choice);
void client (char *fullName, char *friendlyName);
void server (char *fullName, char *friendlyName);

struct Message;
void send (Message *message);
Message *receive ();
                                     
char *playerList ();
void choosePlayer (long choice);


//***********************************************************************//
//               Private Prototypes and Forward References               //
//***********************************************************************//

void noteError (char *message);
void resetError ();
Boolean errorExists ();

void reallocateReceiveBuffer (long newSize);

struct ServiceProvider;
void closeServiceProviders ();
void closeServiceProvider (ServiceProvider *serviceProvider);

void closeSessions ();

struct Player {
	char *formalName, *friendlyName;
	DPID identifier;
	Player *next;
};
Player *activeReceivingPlayer; 
void closePlayers ();
void closePlayer (Player *player);

//***********************************************************************//
//                                Tools                                  //
//***********************************************************************//

typedef BYTE Boolean;

#define true 1
#define false 0

#define run(code,message) \
	{HRESULT result = code; \
	if (result != DP_OK) { \
		noteError (message); \
	}}

#define runClean(code,message) \
	{HRESULT result = code; \
	if (result != DP_OK && result != DPERR_NOSESSIONS) { \
		noteError (message); \
	}}
 
#define worked(code) (code == DP_OK)

#define space(variable,type,returnResult) \
	variable = new type; \
	if (variable == NULL) { \
		noteError ("Out of space!"); \
		return returnResult; \
	}

//***********************************************************************//
//                             Error Processing                          //
//***********************************************************************//

#define setupStaticBuffer() \
	static char buffer [500]; va_list parameters; \
	va_start (parameters, message); \
	vsprintf (buffer, message, parameters); \
	va_end (parameters) 

char *asString (char *message, ...) { 
	setupStaticBuffer ();
	return (char *) &buffer;	 				 
}	

void trace (char *message, ...) { 
	setupStaticBuffer ();
	MessageBox (NULL, (char *) &buffer, "Tracing", MB_OK);
}

char *errorMessage;

char *lastError () {
	return errorMessage;
}

void noteError (char *message) {
	errorMessage = message;
}

void resetError () {
	errorMessage = "";
}

Boolean errorExists () {
	return *errorMessage != '\0';
}

//***********************************************************************//
//                          The Direct Play Object                       //
//***********************************************************************//

typedef IDirectPlay DirectPlay;
                                            
DirectPlay *directPlay = NULL;
HANDLE directPlayEvent = NULL; //We don't use triggered notification.

GUID guid;

void openDirectPlay (char *guidString) {
	//A guid string is a double byte strings such as
	//"{d4ba6b60-00d7-11d0-88d2-00a024a57dd9}"obtained 
	//by running c:\msdev\bin\uuidgen.exe.
	closeDirectPlay ();
	resetError ();
	run (CLSIDFromString ((LPOLESTR) guidString, &guid), "Invalid GUID");
	if (errorExists ()) return;
	activeReceivingPlayer = NULL;
	reallocateReceiveBuffer (400);
}

void closeDirectPlay () {
	resetError ();
	closeServiceProviders ();
	closeSessions ();
	closePlayers ();
	if (directPlay != NULL) {
		directPlay->Close (); directPlay->Release (); directPlay = NULL;
	}
}

//***********************************************************************//
//                             Service Providers                         //
//***********************************************************************//
                                             
struct ServiceProvider {
	GUID driverGuid; 
	char *driverName;
	long majorVersion, minorVersion;
	ServiceProvider *next;
};

ServiceProvider *firstServiceProvider, *lastServiceProvider;
ServiceProvider *activeProvider; 

char *smalltalkProviderList = NULL;
long smalltalkProviderListSize;

#define providerLoop(variable,code) \
	{for (ServiceProvider *variable = firstServiceProvider; variable != NULL; variable = variable->next) {\
		code\
	}}

Boolean FAR PASCAL enumerateServiceProviderCallback (GUID *guid, 
	char *driverName, long majorVersion, long minorVersion, void *unused) {
	//Create and link up a new provider.
	ServiceProvider *space (newProvider, ServiceProvider, false);
	if(lastServiceProvider == NULL) 
		firstServiceProvider = lastServiceProvider = newProvider;
	else
		lastServiceProvider->next = newProvider;
	newProvider->next = NULL;
	lastServiceProvider = newProvider;

	//Fill in provider information.
	newProvider->driverGuid = *guid;
	long driverNameSize = strlen (driverName) + 1;
	space (newProvider->driverName, char [driverNameSize], false);
	smalltalkProviderListSize += driverNameSize;
	strcpy (newProvider->driverName, driverName);
	newProvider->majorVersion = majorVersion; 
	newProvider->minorVersion = minorVersion; 
	return true; //Everything worked.
}

char *providerList () {
	//Returns the list of providers; an empty list if an error is encountered.
	//Note: format of list is string1 '\1' string2 '\1' ... '\0' (with '\0'
	//instead of '\1' for the last element).
	if (smalltalkProviderList != NULL) return smalltalkProviderList;
	
	//Get the service providers.
	firstServiceProvider = lastServiceProvider = NULL;
	smalltalkProviderListSize = 0;
	resetError (); 
	DirectPlayEnumerate ((LPDPENUMDPCALLBACK) enumerateServiceProviderCallback, NULL);
	if (errorExists ()) return "";

	//Setup the smalltalk provider list.
	space (smalltalkProviderList, char [smalltalkProviderListSize + 1], "");
	char *list = smalltalkProviderList;
	providerLoop (next, {
		long size = strlen (next->driverName) + 1;
		CopyMemory (list, next->driverName, size);
		list += size; *(list-1) = '\1';
	});
	smalltalkProviderList [smalltalkProviderListSize] = '\0';
	return smalltalkProviderList;
}

void chooseProvider (long choice) {
	resetError ();
	long count = 0;
	providerLoop (next, {
		count++;
		if (count == choice) {
			activeProvider = next; 
			run (
				DirectPlayCreate (&activeProvider->driverGuid, &directPlay, NULL), 
				"Provider not operating");
			return;
		}
	});
	activeProvider = NULL;
	noteError ("Selected provider out of bounds");
}

void closeServiceProviders () {
	//Delete the provider list.
	ServiceProvider *last = NULL;
	providerLoop (next, {
		if (last != NULL) closeServiceProvider (last);
		last = next;
	});
	if (last != NULL) closeServiceProvider (last);
	firstServiceProvider = lastServiceProvider = activeProvider = NULL;
	//Delete the provider string.
	if (smalltalkProviderList != NULL) delete [] smalltalkProviderList;
	smalltalkProviderList = NULL;
}

void closeServiceProvider (ServiceProvider *serviceProvider) {
	//Delete the space for the serviceProvider struct
	delete [] serviceProvider->driverName;
	delete serviceProvider;
}

//***********************************************************************//
//                               Sessions                                //
//***********************************************************************//                             

enum UserType {userIsClient, userIsServer};

UserType userType; DPID userIdentifier;
 
typedef DPSESSIONDESC SessionDescription;
                                       
struct Session {
	SessionDescription description;
	Session *next;
};

Session *firstSession, *lastSession;
Session *activeSession; long activeSessionIndex;

char *smalltalkSessionList;
long smalltalkSessionListSize;

#define sessionLoop(variable,code) \
	{for (Session *variable = firstSession; variable != NULL; variable = variable->next) {\
		code\
	}}

long lastValueOfSmalltalkSessionListSize;

Boolean FAR PASCAL enumerateSessionCallback (SessionDescription *description, 
	void *unused, long *timeout, long flags) {
	//Deal with timing out.
	if (flags & DPESC_TIMEDOUT) {
		if (smalltalkSessionListSize > 0 && 
			smalltalkSessionListSize > lastValueOfSmalltalkSessionListSize) {
			lastValueOfSmalltalkSessionListSize = smalltalkSessionListSize;
			//Managed to get a partial list so far. So continue.
			*timeout += 5000;
			return true;
		}
		return false; //Time to quit.
	}

	//Create and link up a new session.
	Session *space (newSession, Session, false);
	if (lastSession == NULL) {
		firstSession = lastSession = newSession;
		lastValueOfSmalltalkSessionListSize = 0;
	} else
		lastSession->next = newSession;
	newSession->next = NULL;
	lastSession = newSession;

	//Fill in session information.
	newSession->description = *description;
	smalltalkSessionListSize += strlen (description->szSessionName) + 1;
	return true; //Everything worked.
}

char *sessionList (long timeout) {
	//Returns the list of sessions; an empty list if an error is encountered.
	//Note: format of list is string1 '\1' string2 '\1' ... '\0' (with '\0'
	//instead of '\1' for the last element).

	//Release old space if this is not the first time.
	closeSessions ();

	//Create an initial session description to contain the guid.
	SessionDescription description;
	ZeroMemory (&description, sizeof (SessionDescription));
	description.dwSize = sizeof (SessionDescription);
	description.guidSession = guid;
	description.dwFlags = DPENUMSESSIONS_ALL;

	//Get the list of existing sessions.
	firstSession = lastSession = NULL;
	smalltalkSessionListSize = 0;
	resetError (); 
	runClean (directPlay->EnumSessions (&description, timeout, 
		(LPDPENUMSESSIONSCALLBACK) &enumerateSessionCallback, NULL, 
		DPENUMSESSIONS_AVAILABLE), "Session enumeration failed");
	if (errorExists ()) return "";

	//Setup the smalltalk session list.
	space (smalltalkSessionList, char [smalltalkSessionListSize + 1], "");
	char *list = smalltalkSessionList;
	sessionLoop (next, {
		long size = strlen (next->description.szSessionName) + 1;
		CopyMemory (list, next->description.szSessionName, size);
		list += size; *(list-1) = '\1';
	});
	smalltalkSessionList [smalltalkSessionListSize] = '\0';
	return smalltalkSessionList;
}

void chooseSession (long choice) {
	resetError ();
	long count = 0;
	sessionLoop (next, {
		count++;
		if (count == choice) {
			activeSession = next; 
			activeSessionIndex = activeSession->description.dwSession;
			return;
		}
	});
	activeSession = NULL;
	noteError ("Selected session out of bounds");
}

void client (char *fullName, char *friendlyName) {
	//Sets up user as a client and returns true if successful; false, otherwise.
	resetError (); userType = userIsClient;

	//Create an initial session description to contain the guid.
	SessionDescription description;
	ZeroMemory (&description, sizeof (SessionDescription));
	description.dwSize = sizeof (SessionDescription);
	description.guidSession = guid;
	description.dwFlags = DPOPEN_OPENSESSION;
	description.dwSession = activeSessionIndex;
	if (!worked (directPlay->Open (&description))) {
		noteError ("Client could not be set up");
		return;
	}
	if (!worked (directPlay->CreatePlayer (&userIdentifier, friendlyName, fullName, &directPlayEvent))) {
		noteError ("Player could not be created");
		return;
	}
}

void server (char *fullName, char *friendlyName) {
	//Sets up user as a server and returns true if successful; false, otherwise.
	resetError (); userType = userIsServer;

	//Create an initial session description to contain the guid.
	SessionDescription description;
	ZeroMemory (&description, sizeof (SessionDescription));
	description.dwSize = sizeof (SessionDescription);
	description.dwMaxPlayers = 1000000;
	description.guidSession = guid;
	description.dwFlags = DPOPEN_CREATESESSION;
	strcpy (description.szSessionName, fullName);
	if (!worked (directPlay->Open (&description))) {
		noteError ("Server could not be set up");
		return;
	}
	if (!worked (directPlay->CreatePlayer (&userIdentifier, friendlyName, fullName, &directPlayEvent))) {
		noteError ("Player could not be created");
		return;
	}
}

struct Message {
	unsigned long size; //size of data (not of Message).
	BYTE data [1];
};

Message emptyReceiveBuffer = {0, 0};
Message *receiveBuffer = NULL; 
long receiveBufferSizeLimit = 0; //maximum value for size

void reallocateReceiveBuffer (long newSize) {
	//Allocate Message where data length is rounded up to a 
	//multiple of an 'unsigned long'.
	if (receiveBuffer != NULL) 
		delete [] (unsigned long *) receiveBuffer;
	long typeSize = sizeof (unsigned long);
	long longSize = newSize / typeSize +
		1 /*to round up data bytes*/ + 1 /*for size*/;
	receiveBuffer = (Message *) new unsigned long [longSize];
	if (receiveBuffer == NULL) {
		receiveBufferSizeLimit = 0;
		noteError ("Not enough space for receive buffer");
	}
	receiveBufferSizeLimit = longSize * typeSize - typeSize;
}

void send (Message *message) {
	long sender = userIdentifier;
	long receiver = activeReceivingPlayer == NULL ? 0 : activeReceivingPlayer->identifier;
	directPlay->Send (sender, receiver, DPSEND_TRYONCE, message->data, message->size);
}

Message *receive () {
	//Returns the address of a message structure. It is the user's responsibility
	//to copy the data since subsequent messages overwrite the same space.

	//Is there a message?
	unsigned long count;
	run (directPlay->GetMessageCount (userIdentifier, &count), "Unknown player attempting to receive");
	if (count == 0) {
		//No, there is no message.
		return &emptyReceiveBuffer;
	} 

	//Try to get the message.
	unsigned long byteSize = receiveBufferSizeLimit;
	DPID fromIdentifier; DPID toIdentifier = userIdentifier;
	HRESULT result = directPlay->Receive (&fromIdentifier, &toIdentifier, 
		DPRECEIVE_TOPLAYER, &receiveBuffer->data, &byteSize);
	if (result == DPERR_BUFFERTOOSMALL) {
		//No, need more space.
		reallocateReceiveBuffer (byteSize);
		byteSize = receiveBufferSizeLimit;
		//Try again, should work this time.
		result = directPlay->Receive (&fromIdentifier, &toIdentifier, 
			DPRECEIVE_TOPLAYER, &receiveBuffer->data, &byteSize);
	}
	if (result != DP_OK) {
		noteError ("Receive replied with an error");
		return &emptyReceiveBuffer;
	}
	if (fromIdentifier == userIdentifier) return receive (); //Ignore self messages.
	receiveBuffer->size = byteSize;
	return receiveBuffer;
}

void closeSessions () {
	//Delete the session list.
	Session *last = NULL;
	sessionLoop (next, {
		if (last != NULL) delete last;
		last = next;
	});
	if (last != NULL) delete last;
	firstSession = lastSession = activeSession = NULL;
	//Delete the session string.
	if (smalltalkSessionList != NULL) delete [] smalltalkSessionList;
	smalltalkSessionList = NULL;
}

//***********************************************************************//
//                              Players                                  //
//***********************************************************************//                             
                                        
//struct Player {
//	char *formalName, *friendlyName;
//	DPID identifier;
//	Player *next;
//};

Player *firstPlayer, *lastPlayer;
//Player *activeReceivingPlayer; 

char *smalltalkPlayerList;
long smalltalkPlayerListSize;

#define playerLoop(variable,code) \
	{for (Player *variable = firstPlayer; variable != NULL; variable = variable->next) {\
		code \
	}}

Boolean FAR PASCAL enumeratePlayerCallback (DPID identifier, 
	char *formalName, char *friendlyName, long flags, void *unused) {
	//Create and link up a new player.
	Player *space (newPlayer, Player, false);
	if(lastPlayer == NULL) 
		firstPlayer = lastPlayer = newPlayer;
	else
		lastPlayer->next = newPlayer;
	newPlayer->next = NULL;
	lastPlayer = newPlayer;

	//Fill in player information.
	newPlayer->identifier = identifier;
	space (newPlayer->formalName, char [strlen (formalName) + 1], false);
	space (newPlayer->friendlyName, char [strlen (friendlyName) + 1], false);
	smalltalkPlayerListSize += 
		strlen (newPlayer->formalName) + 1 +
		strlen (newPlayer->friendlyName) + 1;
	return true; //Everything worked.
}

char *playerList () {
	//Returns the list of players; an empty list if an error is encountered.
	//Note: format of list is as follows:
	//formalName1 '\1' friendlyName1 '\1' formalName2 '\1' 
	//friendlyName2 '\1' ... '\0' (with '\0'instead of '\1' for the last element).

	//Release old space if this is not the first time.
	closePlayers ();

	//Get the list of existing players.
	firstPlayer = lastPlayer = NULL;
	smalltalkPlayerListSize = 0;
	resetError (); 
	directPlay->EnumPlayers (activeSession->description.dwSession, 
		(LPDPENUMPLAYERSCALLBACK) enumeratePlayerCallback, NULL, 0);
	if (errorExists ()) return "";

	//Setup the smalltalk player list.
	space (smalltalkPlayerList, char [smalltalkPlayerListSize + 1], "");
	//Player list in the form of formalName1, friendlyName1, formalName2, friendlyName2, ..., friendlyNameN.
	char *list = smalltalkPlayerList;
	playerLoop (next, {
		long size = strlen (next->formalName) + 1;
		CopyMemory (list, next->formalName, size);
		list += size; *(list-1) = '\1';
		size = strlen (next->friendlyName) + 1;
		CopyMemory (list, next->friendlyName, size);
		list += size; *(list-1) = '\1';
	});
	smalltalkPlayerList [smalltalkPlayerListSize] = '\0';
	return smalltalkPlayerList;
}

void choosePlayer (long choice) {
	resetError ();
	if (choice == 0) {
		//The "broadcast to everyone" player.
		activeReceivingPlayer = NULL;
		return;
	}
	long count = 0;
	playerLoop (next, {
		count++;
		if (count == choice) {
			activeReceivingPlayer = next; 
			return;
		}
	});
	activeReceivingPlayer = NULL;
	noteError ("Selected player out of bounds");
}

void closePlayers () {
	//Delete the player list.
	Player *last = NULL;
	playerLoop (next, {
		if (last != NULL) closePlayer (next);
		last = next;
	});
	if (last != NULL) closePlayer (last);
	firstPlayer = lastPlayer = activeReceivingPlayer = NULL;
	//Delete the player string.
	if (smalltalkPlayerList != NULL) delete [] smalltalkPlayerList;
	smalltalkPlayerList = NULL;
}

void closePlayer (Player *player) {
	//Delete the space for a player strucuture
	delete [] player->formalName;
	delete [] player->friendlyName;
	delete player;
}